import os
from datetime import datetime, timedelta
from typing import List
from urllib.parse import urlencode
from warnings import warn

import smartcar.config as config
import smartcar.helpers as helpers
import smartcar.types as types


class AuthClient(object):
    def __init__(
        self,
        client_id=None,
        client_secret=None,
        redirect_uri=None,
        test_mode=None,
        mode="live",
    ):
        """
        A client for accessing the Smartcar API.

        NOTE: It is recommended that you set environment variables with your client id, secret, and redirect URI.
        AuthClient by default will search for environment variables "SMARTCAR_CLIENT_ID",
        "SMARTCAR_CLIENT_SECRET", and "SMARTCAR_REDIRECT_URL" and set those as attributes.
        These CAN be passed as arguments as you instantiate AuthClient. Any arguments passed in will override
        and take precedence over the corresponding environment variable.

        However, if neither an environment variable nor an argument is passed in, an exception will be raised.

        Args:
            client_id (str, optional): The application id, provided in the application
                dashboard

            client_secret (str, optional): The application secret, provided in the
                application dashboard

            redirect_uri (str, optional): The URL to redirect to after the user accepts
                or declines the application's permissions. This URL must also be
                present in the Redirect URIs field in the application dashboard

            test_mode (bool, optional): Deprecated, please use `mode` instead.
                Launch Smartcar Connect in [test mode](https://smartcar.com/docs/guides/testing/).

            mode (str, optional): Determine what mode Smartcar Connect should be launched in.
                Should be one of test, live or simulated. Defaults to live.
        """
        self.client_id = client_id or os.environ.get("SMARTCAR_CLIENT_ID")
        self.client_secret = client_secret or os.environ.get("SMARTCAR_CLIENT_SECRET")
        self.redirect_uri = redirect_uri or os.environ.get("SMARTCAR_REDIRECT_URI")
        self.mode = mode.lower()

        if test_mode is not None:
            warn(
                'The "testMode" parameter is deprecated, please use the "mode" parameter instead.',
                DeprecationWarning,
            )
            self.mode = "test" if test_mode else "live"

        self.auth = (self.client_id, self.client_secret)

        if (
            self.client_id is None
            or self.client_secret is None
            or self.redirect_uri is None
        ):
            raise Exception(
                "AuthClient MUST have client_id, client_secret, and redirect_uri attributes."
                "Either set these as environment variables, OR pass them in as arguments when instantiating "
                "AuthClient. The recommended course of action is to set up environment variables "
                "with your client credentials. i.e.: "
                "'SMARTCAR_CLIENT_ID', 'SMARTCAR_CLIENT_SECRET', and 'SMARTCAR_REDIRECT_URI'"
            )
        if self.mode not in ["test", "live", "simulated"]:
            raise Exception(
                "The \"mode\" parameter MUST be one of the following: 'test', 'live', 'simulated'",
            )

    def get_auth_url(self, scope: List[str], options: dict = None) -> str:
        """
        Generate the Connect URL

        Args:
            scope (str[], required): A list of permissions requested by the application

            options (dict, optional): Can have the following keys:

                force_prompt (bool, optional): Set to True in order to force the approval
                    dialog shown to the user. Defaults to False.

                state (str, optional): A random string that will be passed back on
                    redirect, this allows protection against cross-site forgery
                    requests. Defaults to None.

                make_bypass (str, optional): A string that represents a make(car brand).
                    Allows users to bypass the car brand selection screen, allowing the
                    user to go directly to the vehicle login screen. Defaults to None.

                single_select (dictionary, optional): An optional value that
                    sets the behavior of the grant dialog displayed to the user.
                    Can have keys of enabled(bool) and vin(str).
                    if options['enabled'] == True, `single_select`
                    limits the user to selecting only one vehicle. If `single_select` contains a property
                    of `vin`, Smartcar will only authorize the vehicle with the specified VIN.
                    See the [Single Select guide](https://smartcar.com/docs/guides/single-select/)
                    for more information. Defaults to None.

                flags: dictionary(str, bool): An optional list of feature flags that your
                    application has early access to.

                user: (str, optional): An optional developer-defined unique identifier for a vehicle owner.
                    This ID is used to track and aggregate analytics across Connect sessions for each
                    vehicle owner.

        Returns:
            str: authorization url

        Raises:
            SmartcarException
        """
        base_url = config.CONNECT_URL

        query = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "approval_prompt": "auto",
            "scope": " ".join(scope),
            "mode": self.mode,
        }

        if options:
            if options.get("force_prompt"):
                query["approval_prompt"] = "force"

            if options.get("state"):
                query["state"] = options["state"]

            if options.get("make_bypass"):
                query["make"] = options["make_bypass"]

            if options.get("single_select"):
                single_select = options["single_select"]

                if single_select.get("vin"):
                    query["single_select_vin"] = single_select["vin"]
                    query["single_select"] = "true"
                elif single_select.get("enabled"):
                    query["single_select"] = "true"
                else:
                    query["single_select"] = "false"

            if options.get("flags"):
                flags_str = helpers.format_flag_query(options["flags"])
                query["flags"] = flags_str

            if options.get("user"):
                query["user"] = options["user"]

        return base_url + "/oauth/authorize?" + urlencode(query)

    def exchange_code(self, code: str, options: dict = None) -> types.Access:
        """
        Exchange an authentication code for an access dictionary

        Args:
            code (str): A valid authorization code

            options (dict, optional): Can have the following keys:

                flags: dictionary(str, bool): An optional list of feature flags that your
                    application has early access to.

        Returns:
            Access: containing access information, including access_token and
                refresh_token

        Raises:
            SmartcarException
        """
        method = "POST"
        url = config.AUTH_URL
        data = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.redirect_uri,
        }
        params = {}

        if options:
            if options.get("flags"):
                flags_str = helpers.format_flag_query(options["flags"])
                params["flags"] = flags_str

        response = helpers.requester(
            method, url, data=data, auth=self.auth, params=params
        )
        data = response.json()
        return types.make_access_object(_set_expiration(data))

    def exchange_refresh_token(
        self, refresh_token: str, options: dict = None
    ) -> types.Access:
        """
        Exchange a refresh token for a new access dictionary

        Args:
            refresh_token (str): A valid refresh token from a previously retrieved
                access object

            options (dict, optional): Can have the following keys:
                flags: dictionary(str, bool): An optional list of feature flags that your
                    application has early access to.

        Returns:
            Access: containing access information, including access_token and
                refresh_token

        Raises:
            SmartcarException
        """
        method = "POST"
        url = config.AUTH_URL
        data = {"grant_type": "refresh_token", "refresh_token": refresh_token}
        params = {}

        if options:
            if options.get("flags"):
                flags_str = helpers.format_flag_query(options["flags"])
                params["flags"] = flags_str

        response = helpers.requester(
            method, url, data=data, auth=self.auth, params=params
        )
        data = response.json()
        return types.make_access_object(_set_expiration(data))


# Static helpers for AuthClient


def _set_expiration(access: dict) -> dict:
    expire_date = datetime.utcnow() + timedelta(seconds=access["expires_in"])
    refresh_expire_date = datetime.utcnow() + timedelta(days=60)
    access["expiration"] = expire_date
    access["refresh_expiration"] = refresh_expire_date
    return access



from collections import namedtuple
import json
from typing import List
import smartcar.config as config
import smartcar.helpers as helpers
import smartcar.smartcar
import smartcar.types as types
import smartcar.exception as sce


class Vehicle(object):
    def __init__(self, vehicle_id: str, access_token: str, options: dict = None):
        """
        Initializes a new Vehicle to use for making requests to the Smartcar API.

        Args:
            vehicle_id (str): the vehicle's unique identifier

            access_token (str): a valid access token

            options (dict, optional): Can contain the following keys:
                unit_system (str, optional): the unit system to use for vehicle data.
                    Defaults to metric.

                version(str, optional): Version of Smartcar API an instance of vehicle
                    will send requests to. This will override the instance's base url attribute.

                flags(dict, optional): Object of flags where key is the name of the flag and
                    value is string or boolean value.

        Attributes:
            self.vehicle_id (str)
            self.access_token (str): Access token retrieved from Smartcar Connect
            self.api_version (str): e.g. "1.0" or "2.0"
            self.unit_system (str): Must be "metric" or "imperial" (case insensitive)
        """
        self.vehicle_id = vehicle_id
        self.access_token = access_token
        self._api_version = smartcar.smartcar.API_VERSION
        self._unit_system = "metric"
        self._flags = {}

        if options:
            if options.get("unit_system"):
                self._unit_system = options["unit_system"]

            if options.get("version"):
                self._api_version = options["version"]

            if options.get("flags"):
                self._flags = options["flags"]

    def vin(self) -> types.Vin:
        """
        GET Vehicle.vin

        Returns:
            Vin: NamedTuple("Vin", [("vin", str), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "vin"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def charge(self) -> types.Charge:
        """
        GET Vehicle.charge

        Returns:
            Charge: NamedTuple("Charge", [("is_plugged_in", bool), ("state", str), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "charge"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def battery(self) -> types.Battery:
        """
        GET Vehicle.battery

        Returns:
            Battery: NamedTuple("Battery", [("percent_remaining", float),
                ("range", float),
                ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "battery"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def battery_capacity(self) -> types.BatteryCapacity:
        """
        GET Vehicle.battery_capacity

        Returns:
            BatteryCapacity: NamedTuple("BatteryCapacity", [("capacity", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "battery/capacity"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def fuel(self) -> types.Fuel:
        """
        GET Vehicle.fuel

        Returns:
            Fuel: NamedTuple("Fuel", [("range", float),
                ("percent_remaining", float), ("amount_remaining", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "fuel"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def tire_pressure(self) -> types.TirePressure:
        """
        GET Vehicle.tire_pressure

        Returns:
            TirePressure: NamedTuple("tirePressure", [
                ("front_left", int), ("front_right", int), ("back_left", int), ("back_right", int),
                ("meta", rs.namedtuple)
                ])

        Raises:
            SmartcarException
        """
        path = "tires/pressure"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def engine_oil(self) -> types.EngineOil:
        """
        GET Vehicle.engine_oil

        Returns:
            EngineOil: NamedTuple("EngineOil", [("life_remaining", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "engine/oil"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def odometer(self) -> types.Odometer:
        """
        GET Vehicle.odometer

        Returns:
            Odometer: NamedTuple("Odometer", [("distance", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "odometer"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def location(self) -> types.Location:
        """
        GET Vehicle.location

        Returns:
            Location: NamedTuple("Location", [("latitude", float), ("longitude", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "location"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def permissions(self, paging: dict = None):
        """
        GET Vehicle.permissions

        Args:
            paging (dict, optional): Can contain "limit" or "offset":
                limit (int, optional): The number of permissions to return

                offset (int, optional): The index to start permission list at

        Returns:
            list: vehicle's permissions

        Raises:
            SmartcarException
        """
        path = "permissions"
        url = self._format_url(path)
        headers = self._get_headers()

        if paging is None:
            response = helpers.requester("GET", url, headers=self._get_headers())
        else:
            limit = paging.get("limit", 25)
            offset = paging.get("offset", 0)
            response = helpers.requester(
                "GET", url, headers=headers, params={"limit": limit, "offset": offset}
            )

        return types.select_named_tuple(path, response)

    def attributes(self) -> types.Attributes:
        """
        GET Vehicle.attributes

        Returns:
            Attributes: NamedTuple("Attributes", [("id", str), ("make", str), ("model", str), ("year", str),
            ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = ""
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def get_charge_limit(self) -> types.ChargeLimit:
        """
        GET Vehicle.get_charge_limit

        Returns:
            ChargeLimit = NamedTuple("ChargeLimit", [("limit", float), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        path = "charge/limit"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    def lock_status(self) -> types.LockStatus:
        """
            GET Vehicle.lock_status

            Returns:
                LockStatus = NamedTuple("LockStatus", [
                    ("is_locked", bool)
                    ("doors", List[Door]),
                    ("windows", List[Window]),
                    ("sunroof", List[Sunroof]),
                    ("storage", List[Storage]),
                    ("charging_port", List[ChargingPort]),
                    ("meta", namedtuple)
                ]
        )
            Raises:
                SmartcarException
        """
        path = "security"
        url = self._format_url(path)
        headers = self._get_headers()
        response = helpers.requester("GET", url, headers=headers)
        return types.select_named_tuple(path, response)

    # ===========================================
    # Action (POST) Requests
    # ===========================================

    def lock(self) -> types.Status:
        """
        POST Vehicle.lock

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("security")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST", url, headers=headers, json={"action": "LOCK"}
        )
        return types.select_named_tuple("lock", response)

    def unlock(self) -> types.Status:
        """
        POST Vehicle.unlock

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("security")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST", url, headers=headers, json={"action": "UNLOCK"}
        )
        return types.select_named_tuple("unlock", response)

    def start_charge(self) -> types.Status:
        """
        POST Vehicle.start_charge

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("charge")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST", url, headers=headers, json={"action": "START"}
        )
        return types.select_named_tuple("start_charge", response)

    def stop_charge(self) -> types.Status:
        """
        POST Vehicle.stop_charge

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("charge")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST", url, headers=headers, json={"action": "STOP"}
        )
        return types.select_named_tuple("stop_charge", response)

    def set_charge_limit(self, limit) -> types.Status:
        """
        POST Vehicle.set_charge_limit

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("charge/limit")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST", url, headers=headers, json={"limit": limit}
        )
        return types.select_named_tuple("set_charge_limit", response)

    def send_destination(self, latitude, longitude) -> types.Action:
        """
        POST Vehicle.send_destination

        Returns:
            Action: NamedTuple("Action", [("status", str), ("message", str), ("meta", rs.namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("navigation/destination")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester(
            "POST",
            url,
            headers=headers,
            json={"latitude": latitude, "longitude": longitude},
        )
        return types.select_named_tuple("send_destination", response)

    def batch(self, paths: List[str]) -> namedtuple:
        """
        POST Vehicle.batch

        This method follows a series of steps:
        1. Format and send request to Smartcar API batch endpoint
        2. Store each response in a batch dictionary.
        3. Attach meta object for the high-level batch request to the dictionary
        4. Transform batch_dict into a namedtuple, and return

        Args:
            paths (str[]): an array of paths to make
            the batch request to

        Returns:
            namedtuple: the responses from Smartcar API, each attribute is a lambda
            that returns the appropriate NamedTuple OR raises a SmartcarException (if
            the request results in an error).

        Raises:
            SmartcarException
        """
        # STEP 1 - Send Request
        url = self._format_url("batch")
        headers = self._get_headers()
        json_body = {"requests": [{"path": path} for path in paths]}
        response = helpers.requester("POST", url, headers=headers, json=json_body)

        # STEP 2 - Format batch_dict
        # [KEYS] will represent the path sent in the batch.
        # The name of the key will eventually be the name of the method attached to the final return.
        # [VALUES] are lambdas that return a NamedTuple OR raises a SmartcarException, depending on the
        # success of the request.
        batch_dict = dict()
        path_responses = response.json()["responses"]
        for res_dict in path_responses:
            path, attribute = helpers.format_path_and_attribute_for_batch(
                res_dict["path"]
            )

            if res_dict.get("code") == 200:
                # attach top-level sc-request-id to res_dict
                res_dict["headers"]["sc-request-id"] = response.headers.get(
                    "sc-request-id"
                )
                # use lambda default args to avoid issues with closures
                batch_dict[attribute] = (
                    lambda p=path, r=res_dict: types.select_named_tuple(p, r)
                )
            else:
                # if individual response is erroneous, attach a lambda that returns a SmartcarException
                def _attribute_raise_exception(smartcar_exception):
                    raise smartcar_exception

                code = res_dict.get("code")
                headers = response.headers
                body = json.dumps(res_dict.get("body"))
                sc_exception = sce.exception_factory(code, headers, body)
                batch_dict[attribute] = (
                    lambda e=sc_exception: _attribute_raise_exception(e)
                )

        # STEP 3 - Attach Meta to batch_dict
        batch_dict["meta"] = types.build_meta(response.headers)

        # STEP 4 - Transform batch_dict into a namedtuple
        return types.generate_named_tuple(batch_dict, "batch")

    # ===========================================
    # DELETE requests
    # ===========================================

    def disconnect(self) -> types.Status:
        """
        Disconnect this vehicle from the connected application.

        Note: Calling this method will invalidate your access token and you will
        have to have the user reauthorize the vehicle to your application if you
        wish to make requests to it

        Returns:
            Status: NamedTuple("Status", [("status", str), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url("application")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester("DELETE", url, headers=headers)
        return types.select_named_tuple("disconnect", response)

    # ===========================================
    # Webhook requests
    # ===========================================

    def subscribe(self, webhook_id: str) -> types.Subscribe:
        """
        Subscribe a vehicle to a webhook

        Args:
            webhook_id (str)

        Returns:
            Subscribe: NamedTuple("Subscribe", [("webhook_id", str"), ("vehicle_id", str), ("meta", namedtuple)
        """
        url = self._format_url(f"webhooks/{webhook_id}")
        headers = self._get_headers(need_unit_system=False)
        response = helpers.requester("POST", url, headers=headers)
        return types.select_named_tuple("subscribe", response)

    def unsubscribe(self, amt: str, webhook_id: str) -> types.Status:
        """
        Subscribe a vehicle to a webhook

        Args:
            amt(str): Application Management Token, found on Smartcar Dashboard
            webhook_id (str)

        Returns:
            Status: NamedTuple("Subscribe", [("webhook_id", str"), ("vehicle_id", str), ("meta", namedtuple)
        """
        url = self._format_url(f"webhooks/{webhook_id}")

        # Note: Authorization header is different, compared to the other methods
        headers = {"Authorization": f"Bearer {amt}"}
        response = helpers.requester("DELETE", url, headers=headers)
        return types.select_named_tuple("unsubscribe", response)

    # ===========================================
    # General Purpose Request Method
    # ===========================================

    def request(
        self, method: str, path: str, body: dict = {}, headers: dict = {}
    ) -> types.Response:
        """
        Utility method to make a request to a Smartcar endpoint - can be used
        to make requests to brand specific endpoints.

        Args:
            method (str): The HTTP request method to use.
            path (str): The path to make the request to.
            body (dict): The request body.
            headers (dict): The headers to include in the request.

        Returns:
            Response = NamedTuple("Response", [("body", dict), ("meta", namedtuple)])

        Raises:
            SmartcarException
        """
        url = self._format_url(path)

        # Authorization header not provided
        if not "Authorization" in headers:
            has_units_header = "sc-unit-system" in headers
            generated_headers = self._get_headers(
                need_unit_system=(not has_units_header)
            )
            headers.update(generated_headers)

        response = helpers.requester(method, url, headers=headers, json=body)

        return types.select_named_tuple("request", response)

    # ===========================================
    # Utility
    # ===========================================
    def set_unit_system(self, unit_system: str) -> None:
        """
        Change unit system of this instance of smartcar.Vehicle

        Args:
            unit_system (str): Must be "metric" or "imperial",
                case insensitive.

        Returns: None
        """
        if unit_system.lower() != "metric" or unit_system.lower() != "imperial":
            raise ValueError(
                "'unit_system' must be either 'metric' or 'imperial' (case insensitive)"
            )

        else:
            self._unit_system = unit_system.lower()

    # ===========================================
    # Private methods
    # ===========================================
    def _format_query_params(self) -> str:
        """
        Returns (str): Query parameters as a query string
        """
        if self._flags:
            return f"?flags={helpers.format_flag_query(self._flags)}"
        return ""

    def _format_url(self, path: str) -> str:
        """
        Returns (str): Base url with current API version.
        User can change api_version attribute at will.
        """
        return f"{config.API_URL}/v{self._api_version}/vehicles/{self.vehicle_id}/{path}{self._format_query_params()}"

    def _get_headers(self, need_unit_system: bool = True) -> dict:
        """
        Returns (dict): Provides 'Authorization' Header
        and (optionally) the 'sc-unit-system' header.
        Some endpoints don't require the 'sc-unit-system' header.
        """
        headers = {"Authorization": f"Bearer {self.access_token}"}

        if need_unit_system:
            headers["sc-unit-system"] = self._unit_system.lower()

        return headers



import json


class SmartcarException(Exception):
    """
    All exceptions with this SDK will be a Smartcar Exception. v1.0 errors and
    v2.0 errors are distinguished by the response body. "General" exceptions will
    be raised as "SDK ERRORS".
    """

    def __init__(self, **kwargs):
        # populate the error with passed in keys.
        for key in kwargs.keys():
            # format resolution: must be a dictionary with "type" and "url"
            if key == "resolution":
                if type((kwargs[key])) == str:
                    formatted_resolution = {"type": kwargs[key], "url": None}
                    self.__dict__[key] = formatted_resolution
                if type(kwargs[key]) == dict:
                    formatted_resolution = {
                        "type": kwargs[key].get("type"),
                        "url": kwargs[key].get("url"),
                    }
                    self.__dict__[key] = formatted_resolution
                else:
                    self.__dict__[key] = {"type": None, "url": None}
            else:
                self.__dict__[key] = kwargs[key] or None

        # v2.0
        if "type" in kwargs and "code" in kwargs and "description" in kwargs:
            code = f":{kwargs.get('code')}" if kwargs.get("code") else ""
            self.message = f"{kwargs['type']}{code} - {kwargs['description']}"

        # v1.0
        else:
            self.message = kwargs.get("message")

        super().__init__(self.message)


def exception_factory(status_code: int, headers: dict, body: str):
    # v1.0 Exception: Content type other than application/json
    if "application/json" not in headers["Content-Type"]:
        return SmartcarException(status_code=status_code, message=body)

    # Parse body into JSON. Throw SDK error if this fails.
    try:
        response = json.loads(body)
    except Exception:
        return SmartcarException(
            status_code=status_code,
            request_id=headers.get("SC-Request-Id"),
            type="SDK_ERROR",
            message=body,
        )

    # v1.0 with code or OAuth error
    if response.get("error"):
        if response.get("message"):
            # smartcar v1
            return SmartcarException(
                status_code=response.get("statusCode"),
                request_id=response.get("requestId"),
                type=response.get("error"),
                message=response.get("message"),
                code=response.get("code"),
            )
        elif response.get("error_description"):
            # OAuth error
            return SmartcarException(
                status_code=status_code,
                type=response.get("error"),
                message=response.get("error_description"),
            )

    # v2.0
    elif response.get("type"):
        # Set retry_after only for Vehicle Rate Limit errors
        if headers.get("Retry-After") is not None:
            return SmartcarException(
                status_code=response.get("statusCode"),
                request_id=response.get("requestId"),
                type=response.get("type"),
                description=response.get("description"),
                code=response.get("code"),
                doc_url=response.get("docURL"),
                resolution=response.get("resolution"),
                detail=response.get("detail"),
                retry_after=headers.get("Retry-After"),
                suggested_user_message=response.get("suggestedUserMessage"),
            )
        else:
            return SmartcarException(
                status_code=response.get("statusCode"),
                request_id=response.get("requestId"),
                type=response.get("type"),
                description=response.get("description"),
                code=response.get("code"),
                doc_url=response.get("docURL"),
                resolution=response.get("resolution"),
                detail=response.get("detail"),
                suggested_user_message=response.get("suggestedUserMessage"),
            )

    # Weird...
    else:
        return SmartcarException(
            status_code=status_code,
            request_id=headers["SC-Request-Id"],
            type="SDK_ERROR",
            message=body,
        )



__version__ = "6.13.0"

from smartcar.auth_client import AuthClient

from smartcar.exception import SmartcarException

from smartcar.smartcar import (
    get_user,
    get_vehicles,
    get_compatibility,
    hash_challenge,
    get_api_version,
    set_api_version,
    verify_payload,
    get_connections,
    delete_connections,
)

from smartcar.vehicle import Vehicle
